/**
 * I2C RGB Led driver
 * For ATtiny45 @ 8MHz (no divide clock by 8)
 * 
 * www.projects.adamh.cz/I2cRgbLed  
 * 
 * (c) 2010 Adam Heinrich, www.adamh.cz
 *   
 */   
#define F_CPU 8000000  // 8 MHz

#define __ATtiny45__

#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "i2c-slave.c"

/**
 * PIN Definitions
 */ 
#define RGB_DDR() DDRB |= (1<<PB1)|(1<<PB3)|(1<<PB4)
#define R_OFF() PORTB |= (1<<PB1)
#define R_ON() PORTB &= ~(1<<PB1)

#define G_OFF() PORTB |= (1<<PB4)
#define G_ON() PORTB &= ~(1<<PB4)

#define B_OFF() PORTB |= (1<<PB3)
#define B_ON() PORTB &= ~(1<<PB3)


/**
 * I2C settings
 */ 
#define I2C_ADDR 0x10
#define I2C_SET_COLOR_REGISTER 0x01
#define I2C_FADE_TO_COLOR_REGISTER 0x02

/**
 * PWM Control
 */ 
volatile unsigned char red = 0;
volatile unsigned char green = 0;
volatile unsigned char blue = 0;
volatile unsigned char timerCounter = 0;
ISR (TIM0_OVF_vect) {
  if (timerCounter <= red) {
    R_ON();
  } else {
    R_OFF();
  }
  
  if (timerCounter <= green) {
    G_ON();
  } else {
    G_OFF();
  }
  
  if (timerCounter <= blue) {
    B_ON();
  } else {
    B_OFF();
  }
  
  if (timerCounter < 255) {
    timerCounter++;
  } else {
    timerCounter = 1;
  }
}

/**
 * Delay ms
 */ 
void delay_ms(int ms) {
  while (ms) {
    _delay_ms(1);
    ms--;
  }
}

int main(void) {
	// Configure I/O
  RGB_DDR();
  
  R_OFF();
  G_OFF();
  B_OFF();
	
	// I2C init
  i2c_initialize(I2C_ADDR);
  
  // Init Timer/Counter 0
  TCCR0B = (0<<CS02)|(0<<CS01)|(1<<CS00); // CLK/256
  TIMSK = (1<<TOIE0);

  // Enable interrupts
  sei();
  
  // Startup blink
	red = 150;
	_delay_ms(50);
	red = 0;
	_delay_ms(50);
	red = 150;
	_delay_ms(50);
	red = 0;
  
  unsigned char fading = 0;
  unsigned char fadeSpeed = 0;
  
  unsigned char redFade = 0;
  unsigned char greenFade = 0;
  unsigned char blueFade = 0;

  while (1) {  
    // Reply
    if (i2c_reply_ready()) {
      i2c_rdbuf[0] = 0x00;
      i2c_reply_done(1);
    }
      	
    // Message received
    if (i2c_message_ready()) {
      if (i2c_wrlen >= 3) {
        if (i2c_wrbuf[1] == I2C_SET_COLOR_REGISTER) {
          red = i2c_wrbuf[2];
          green = i2c_wrbuf[3];
          blue = i2c_wrbuf[4];
          
          // No fading
          redFade = red;
          greenFade = green;
          blueFade = blue;
          fadeSpeed = 0;
        } else if (i2c_wrbuf[1] == I2C_FADE_TO_COLOR_REGISTER) {
          redFade = i2c_wrbuf[2];
          greenFade = i2c_wrbuf[3];
          blueFade = i2c_wrbuf[4];
          fadeSpeed = i2c_wrbuf[5];
        }
      }
          
      i2c_message_done();                
    }
    
    // Fade
    fading = 0;    
    if (red > redFade) {
      red--;
      fading = 1;
    } else if (red < redFade) {
      red++;
      fading = 1;
    }
    
    if (green > greenFade) {
      green--;
      fading = 1;
    } else if (green < greenFade) {
      green++;
      fading = 1;
    }
    
    if (blue > blueFade) {
      blue--;
      fading = 1;
    } else if (blue < blueFade) {
      blue++;
      fading = 1;
    }
    
    if (fading == 1) {
      delay_ms(fadeSpeed);
    }
  }

	return 0;
}

